"""
pyroaddon - A monkeypatcher add-on for Pyrogram
Copyright (C) 2022 - <https://github.com/pyroaddon>

This file is part of pyroaddon.

pyroaddon is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

pyroaddon is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with pyroaddon.  If not, see <https://www.gnu.org/licenses/>.
"""

import asyncio
import functools
import pyrogram
import typing

from ..utils import patch, patchable

loop = asyncio.get_event_loop()
    
class ListenerCanceled(Exception):
    pass
pyrogram.errors.ListenerCanceled = ListenerCanceled

class Client():
    @patchable
    def __init__(self, *args, **kwargs):
        self.listening = {}
        self.using_mod = True
        
        self.old__init__(*args, **kwargs)
    
    @patchable
    async def listen(self, chat_id: typing.Union[int, str], filters=None, timeout=None):
        if type(chat_id) != int:
            chat = await self.get_chat(chat_id)
            chat_id = chat.id
        
        future = loop.create_future()
        future.add_done_callback(
            functools.partial(self.clear_listener, chat_id)
        )
        self.listening.update({
            chat_id: {"future": future, "filters": filters}
        })
        return await asyncio.wait_for(future, timeout)
    
    @patchable
    async def get_all_groups(self) -> typing.List[pyrogram.types.Chat]:
        chats = pyrogram.types.List()
        for chat in (await self.send(pyrogram.raw.functions.messages.GetAllChats(except_ids=[0]))).chats:
            if isinstance(chat, pyrogram.raw.types.Channel):
                if any(getattr(chat, i, False) for i in ('megagroup', 'gigagroup')):
                    chats.append(Chat(
                        client = self, 
                        id=pyrogram.utils.MAX_CHANNEL_ID - chat.id, 
                        type='supergroup' if chat.megagroup else 
                        'group' if chat.gigagroup else None,
                        title=chat.title,
                        has_protected_content=chat.noforwards,
                        username=chat.username,
                    ))
        return chats
    
    @patchable
    async def get_chat_administrators(self, chat_id: typing.Union[int, str], has_creator: bool=False) -> typing.List[pyrogram.types.User]:
        administrators = pyrogram.types.List()
        for administrator in await self.get_chat_members(chat_id, filter="administrators"):
            if has_creator:
                administrators.append(administrator)
            elif administrator.status != 'creator':
                administrators.append(administrator)
        return administrators
    
    @patchable
    async def ask(
            self, 
            chat_id: typing.Union[int, str], 
            text: str, 
            filters: pyrogram.filters.Filter=None, 
            timeout: str=None, 
            *args, 
            **kwargs
        ):
        request = await self.send_message(chat_id, text, *args, **kwargs)
        response = await self.listen(chat_id, filters, timeout)
        response.request = request
        return response
    
    @patchable
    def clear_listener(self, chat_id: typing.Union[int, str], future):
        if future == self.listening[chat_id]["future"]:
            self.listening.pop(chat_id, None)
    
    @patchable
    def cancel_listener(self, chat_id: typing.Union[int, str]):
        listener = self.listening.get(chat_id)
        if not listener or listener['future'].done():
            return
        
        listener['future'].set_exception(ListenerCanceled())
        self.clear_listener(chat_id, listener['future'])
pyrogram.sync.async_to_sync(Client, 'get_all_groups')
pyrogram.sync.async_to_sync(Client, 'get_chat_administrators')
pyrogram.sync.async_to_sync(Client, 'ask')
pyrogram.sync.async_to_sync(Client, 'listen')
patch(pyrogram.client.Client)(Client)

@patch(pyrogram.handlers.message_handler.MessageHandler)
class MessageHandler():
    @patchable
    def __init__(self, callback: callable, filters=None):
        self.user_callback = callback
        self.old__init__(self.resolve_listener, filters)
    
    @patchable
    async def resolve_listener(self, client: Client, message: pyrogram.types.Message, *args):
        listener = client.listening.get(message.chat.id)
        if listener and not listener['future'].done():
            listener['future'].set_result(message)
        else:
            if listener and listener['future'].done():
                client.clear_listener(message.chat.id, listener['future'])
            await self.user_callback(client, message, *args)
    
    @patchable
    async def check(self, client: Client, update: pyrogram.types.Update):
        listener = client.listening.get(update.chat.id)
        
        if listener and not listener['future'].done():
            return await listener['filters'](client, update) if callable(listener['filters']) else True
            
        return (
            await self.filters(client, update)
            if callable(self.filters)
            else True
        )

@patch(pyrogram.types.user_and_chats.chat.Chat)
class Chat(pyrogram.types.Chat):
    @patchable
    def listen(self, filters=None, timeout=None):
        return self._client.listen(self.id, filters=filters, timeout=timeout)
    
    @patchable
    def get_administrators(self, has_creator: bool=False) -> typing.List[int]:
        return self._client.get_chat_administrators(self.id, has_creator=has_creator)
    
    @patchable
    def ask(self, text: str,  filters: pyrogram.filters.Filter=None, timeout: str=None, *args, **kwargs):
        return self._client.ask(self.id, text=text, filters=filters, timeout=timeout, *args, **kwargs)
    
    @patchable
    def cancel_listener(self):
        return self._client.cancel_listener(self.id)

@patch(pyrogram.types.user_and_chats.user.User)
class User(pyrogram.types.User):
    @patchable
    def listen(self, filters: pyrogram.filters.Filter=None, timeout: str=None):
        return self._client.listen(self.id, filters=filters, timeout=timeout)
    
    @patchable
    def ask(self, text: str,  filters: pyrogram.filters.Filter=None, timeout: str=None, *args, **kwargs):
        return self._client.ask(self.id, text=text, filters=filters, timeout=timeout, *args, **kwargs)
    
    @patchable
    def cancel_listener(self):
        return self._client.cancel_listener(self.id)



"""
pyroaddon - A monkeypatcher add-on for Pyrogram
Copyright (C) 2022 - <https://github.com/pyroaddon>

This file is part of pyroaddon.

pyroaddon is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

pyroaddon is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with pyroaddon.  If not, see <https://www.gnu.org/licenses/>.
"""

import pyrogram 

@pyrogram.filters.create
def dice(_, __, message: pyrogram.types.Message):
    if message.dice:
        return True
    return False
pyrogram.filters.dice = dice

@pyrogram.filters.create
async def ttl_message(_, __, m: pyrogram.types.Message):
    if (m.photo and m.photo.ttl_seconds) or (m.video and m.video.ttl_seconds):
        return True
    return False
pyrogram.filters.ttl_message = ttl_message

@pyrogram.filters.create
async def video_sticker(_, __, m: pyrogram.types.Message):
    if await pyrogram.filters.sticker(__, m):
        return m.sticker.is_video
    return False
pyrogram.filters.video_sticker = video_sticker



"""
pyroaddon - A monkeypatcher add-on for Pyrogram
Copyright (C) 2022 - <https://github.com/pyroaddon>

This file is part of pyroaddon.

pyroaddon is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

pyroaddon is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with pyroaddon.  If not, see <https://www.gnu.org/licenses/>.
"""

import math
from ..helpers import array_chunk

class Pagination:
    def __init__(self, objects, page_data=None, item_data=None, item_title=None):
        default_page_callback = (lambda x: str(x))
        default_item_callback = (lambda i, pg: f'[{pg}] {i}')
        self.objects = objects
        self.page_data = page_data or default_page_callback
        self.item_data = item_data or default_item_callback
        self.item_title = item_title or default_item_callback
    
    def create(self, page, lines=5, columns=1):
        quant_per_page = lines*columns
        page = 1 if page <= 0 else page
        offset = (page-1)*quant_per_page
        stop = offset+quant_per_page
        cutted = self.objects[offset:stop]
        
        total = len(self.objects)
        pages_range = [*range(1, math.ceil(total/quant_per_page)+1)] # each item is a page
        last_page = len(pages_range)
        
        
        nav = []
        if page <= 3:
            for n in [1,2,3]:
                if n not in pages_range:
                    continue
                text = f"· {n} ·" if n == page else n
                nav.append( (text, self.page_data(n)) )
            if last_page >= 4:
                nav.append(
                    ('4 ›' if last_page > 5 else 4, self.page_data(4))
                )
            if last_page > 4:
                nav.append(
                    (f'{last_page} »' if last_page > 5 else last_page, self.page_data(last_page))
                )
        elif page >= last_page-2:
            nav.extend([
                (f'« 1' if last_page-4 > 1 else 1, self.page_data(1)),
                (f'‹ {last_page-3}' if last_page-4 > 1 else last_page-3, self.page_data(last_page-3))
            ])
            for n in range(last_page-2, last_page+1):
                text = f"· {n} ·" if n == page else n
                nav.append( (text, self.page_data(n)) )
        else:
            nav = [
                (f'« 1', self.page_data(1)),
                (f'‹ {page-1}', self.page_data(page-1)),
                (f'· {page} ·', "noop"),
                (f'{page+1} ›', self.page_data(page+1)),
                (f'{last_page} »', self.page_data(last_page)),
            ]
        
        buttons = []
        for item in cutted:
            buttons.append(
                (self.item_title(item, page), self.item_data(item, page))
            )
        kb_lines = array_chunk(buttons, columns)
        if last_page > 1:
            kb_lines.append(nav)
        
        return kb_lines


from pyrogram.types import InlineKeyboardButton, InlineKeyboardMarkup, KeyboardButton, ReplyKeyboardMarkup, ForceReply

def ikb(rows = []):
    lines = []
    for row in rows:
        line = []
        for button in row:
            button = btn(*button) # InlineKeyboardButton
            line.append(button)
        lines.append(line)
    return InlineKeyboardMarkup(inline_keyboard=lines)
    #return {'inline_keyboard': lines}

def btn(text, value, type = 'callback_data'):
    return InlineKeyboardButton(text, **{type: value})
    #return {'text': text, type: value}

# The inverse of above
def bki(keyboard):
    lines = []
    for row in keyboard.inline_keyboard:
        line = []
        for button in row:
            button = ntb(button) # btn() format
            line.append(button)
        lines.append(line)
    return lines
    #return ikb() format

def ntb(button):
    for btn_type in ['callback_data', 'url', 'switch_inline_query', 'switch_inline_query_current_chat', 'callback_game']:
        value = getattr(button, btn_type)
        if value:
            break
    button = [button.text, value]
    if btn_type != 'callback_data':
        button.append(btn_type)
    return button
    #return {'text': text, type: value}

def kb(rows = [], **kwargs):
    lines = []
    for row in rows:
        line = []
        for button in row:
            button_type = type(button)
            if button_type == str:
                button = KeyboardButton(button)
            elif button_type == dict:
                button = KeyboardButton(**button)
            
            line.append(button)
        lines.append(line)
    return ReplyKeyboardMarkup(keyboard=lines, **kwargs)

kbtn = KeyboardButton

def force_reply(selective=True):
    return ForceReply(selective=selective)

def array_chunk(input, size):
    return [input[i:i+size] for i in range(0, len(input), size)]

